
/*-----------------------------------------------------------------*
 *
 * QNX4 NAMPC-II X25 Device Driver
 * by Jean-Claude MICHOT (jcmichot@teaser.fr)
 *
 * (C) 1998 France-Teaser
 *
 * One point worth mentioning is that "most" hardware drivers
 * will provide input into the Device subsystem directly from
 * the input interrupt handler but this X25 driver input data
 * at "process time".
 *
 * On the other hand, it is NOT uncommon for a "real" driver to
 * output data at process time as is the case in this example.
 * (for example Dev.con, Dev.win, Dev.par all output data at
 * process time).
 *
 *-----------------------------------------------------------------*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/sched.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/psinfo.h>
#include <sys/sys_msg.h>
#include <sys/dev.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <process.h>

#include "_devdrvr.h"

extern int far Kick (struct output_buffer far *obuf);
extern int far Stty (struct termios far *tios); 
extern int far Ctrl (union ioctl_entry far *ioc); 

extern int sys_msg( void * );

extern void terminate(int signal);
extern int sys_msg(void *buf);
extern int input_ready(void);
extern void flush_output(void);
extern int init_driver(void);
extern int exit_driver(void);
extern int get_data(int unit,char *buf,int nbytes);
extern int put_data(int unit,char __far *buf,int nbytes,int expand_newlines);
extern void drop_hw(int unit);
extern void raise_hw(int unit);
extern void x25receive(void);
extern void acceptation(int vc,char *data,int lng);
extern void libcv(int vc,int cause,int diagnostique);
extern void echocv(int vc,int on);
extern void dumper(char *buffer,int lng,FILE *fd);

#define MAX_DEVICES 32

union _private_msg {
	msg_t		type;
	msg_t		status;
	char		filer[100];
	} msg;

int idle = 1;
int more = 0;

pid_t out_proxy;
pid_t in_proxy;
pid_t proxy;
                   
#define STX 0x02
#define ETX 0x03
#define cv2tty(n) (n-0x40+1)
#define cv2unit(n) (n-0x40)
#define unit2cv(n) (n+0x40)
#define XQSIZE 1024

int	debug = 0;

int urgent = -1;

enum eurgent {
	URGENT_LIB = 1,
	URGENT_ECHO_ON = 2,
	URGENT_ECHO_OFF = 4
	};

char *devasmt = "/dev/ser1";
char *startprog = "/serv/bin/login";
unsigned old_mode;
int fdasmt;
int	pkt_indisponible,					/* Concentrateur indiponible */
	pkt_cv_saturer	[MAX_DEVICES+8],	/* Saturation d'un CV */
	pkt_cv_utiliser	[MAX_DEVICES+8],
	pkt_cv_echo		[MAX_DEVICES+8],
	pkt_urgent		[MAX_DEVICES+8];

char xbuf[XQSIZE+1];
char xqueue[MAX_DEVICES+1][XQSIZE+1];

//unsigned isize = 8192;
//unsigned osize = 16384;
unsigned isize = 4096;
unsigned osize = 8192;
unsigned csize = 256;

struct driver_ctrl drv_ctrl;
struct device_ctrl ctrl[MAX_DEVICES+1];
int dev_handle;
int num_devices = 32;
unsigned driver_options = DRVR_STTY;  /* found in <sys/_devdrvr.h> */

char *prefix = "ttx";

main(argc, argv)
int argc; char *argv[];
	{
	unsigned c;
	int i;
	pid_t pid;
	int t;
	int nbytes;
	extern void terminate();
	char *fuck;

	setscheduler(0, SCHED_FIFO, 19);

	memset( xqueue, 0, MAX_DEVICES*XQSIZE );
	pkt_indisponible = 0;
	for( i=0; i<MAX_DEVICES; i++ )
		{
		pkt_cv_saturer[i] = pkt_cv_utiliser[i] = 0;
		pkt_cv_echo[i] = 1;
		}

	/*
	 * Allow only SIGTERM, which we catch and cause
	 * graceful termination, all others are ignored
	 */
	for( i = _SIGMIN; i <= _SIGMAX; ++i)
		if( i != SIGTERM)
			signal( i, SIG_IGN );

	signal( SIGTERM, &terminate );

	/*
	 * Scan arguments
	 */
	while ((c=getopt(argc,argv,"C:I:O:N:n:d:t:p:")) != -1) {
		switch (c) {

		case 'C':
			csize = strtol(optarg,NULL,0);
			break;

		case 'I':
			isize = strtol(optarg,NULL,0);
			break;

		case 'O':
			osize = strtol(optarg,NULL,0);
			break;

		case 'N':
			prefix = optarg;
			break;

		case 'd':
			devasmt = optarg;
			break;

		case 'p':
			startprog = optarg;
			break;

		case 't':
			debug = atoi(optarg);
			break;

		case 'n':
			num_devices = strtol(optarg,NULL,0);
			if(num_devices > MAX_DEVICES)
				num_devices = MAX_DEVICES;
			break;

			}
		}

	/*
	 * Allow memory used later, (after dev_drv_register)
	 * this is just to setup the heap pointer
	 */
	fuck = malloc( 128*1024 );

	/*
	 * Register ourselves with the DEVICE Manager
	 */

	if((dev_handle = dev_drvr_register( "x25tty", prefix, num_devices,
					&Kick, &Stty, &Ctrl, &drv_ctrl, driver_options)) == -1) {
		exit(ENFILE);
		}

	free( fuck );

	/* Allocate the resources for this device */

	for(i = 0; i < num_devices; ++i) {
		ctrl[i].isize = isize;
		ctrl[i].osize = osize;
		ctrl[i].csize = csize;
		}

	if(dev_drvr_mount(dev_handle, drv_ctrl.base_tty, num_devices, &ctrl[0])) {
		dev_drvr_shutdown( dev_handle );
		exit(ENOMEM);
		}

	out_proxy = qnx_proxy_attach(0, 0, 0, -1);
	in_proxy = qnx_proxy_attach(0, 0, 0, -1);

	/*
	 * Close off stdin,stdout,stderr since this process
	 * is usually started in sysinit as a background process,
	 * by the shell, which typically sets stdin to /dev/null
	 * (on the machine we booted from), which ties up resources
	 */
	if ( !debug )
		{
		close(0);
		close(1);
		close(2);
		}

	/*
	 * Tell Proc that we will respond to "sin ver" messages
	 */
	qnx_pflags( _PPF_SERVER, _PPF_SERVER, 0, 0 );

	/*
	 * Initialize termios and queue parameters
	 * to whatever defaults make sense for this class of device
	 */
	
	for( i = 0; i < num_devices; ++i ) {
		/* struct input_buffer far *ibuf; */
		struct output_buffer far *obuf;
		struct termios far *tios;

		obuf = ctrl[i].obuf;
		obuf->mode |= OUTPUT_XNL;

		tios = ctrl[i].termios;

		tios->c_ispeed = tios->c_ospeed = 1200;

		tios->c_cflag = (CREAD|HUPCL);
		// Default minitel pas de OPOST
		tios->c_oflag = 0; // (OPOST);
		tios->c_iflag = 0;
		tios->c_lflag = (ISIG|IEXTEN|BRKINT|ECHO);
		tios->c_cc[VEOF]	= 0x00;
		tios->c_cc[VEOL]	= 0x00;
		tios->c_cc[VERASE]	= 0x00;
		tios->c_cc[VINTR]	= 0x00;
		tios->c_cc[VKILL]	= 0x00;
		tios->c_cc[VQUIT]	= 0x00;
		tios->c_cc[VSUSP]	= 0x00;
		tios->c_cc[VSTART]	= 0x00;
		tios->c_cc[VSTOP]	= 0x00;
		tios->c_cc[VMIN]	= 0x00;
		tios->c_cc[VTIME]	= 0x00;
		}

	/* Initialize hardware */

	init_driver();

	/* Let DEV call us now */

	for(i = 0; i < num_devices; ++i) {
		if(dev_drvr_arm(ctrl[i].tty)) {
			dev_drvr_shutdown( dev_handle );
			exit(ENODEV);
			}
		}

	/* Main loop */
	
	for(;;) {
	
		pid = Receive( 0, &msg, sizeof(msg) );

		if(pid == out_proxy) {
			idle = 1;

			/* Output data now available */
			flush_output();
			continue;
			}

		if(pid == in_proxy) {
			/*
			 * Input data is available
			 * NOTE: that this would "normally" be taken care of
			 *       by an interrupt handler
			 */
			input_ready();
			continue;
			}

		/* Otherwise, process a message from a process */

		t = msg.type;
		msg.status = ENOSYS;
		nbytes = sizeof(msg.status);

		switch(t) {

		case 0:
			nbytes = sys_msg(&msg);
			break;

		default:
			break;
			}

		if(nbytes)
			Reply(pid, &msg, nbytes);

		continue;		
		}
	}

void terminate( int signal )
	{
	if(signal) {
		/*
		 * Shutdown by signal
		 */
		dev_drvr_shutdown( dev_handle );
		/*
		 * We should give back any GDT's, etc. that
		 * the system doesn't release for us
		 */
		exit_driver();
		}
	else
		{
		dev_drvr_shutdown( dev_handle );
		exit_driver();
		}
		
	exit(0);
	}

int sys_msg(void *buf) {
	union _sysmsg {
		int status;
		struct _ver {
			struct _sysmsg_hdr hdr;
			struct _sysmsg_version version;
			} ver;
		struct _ver_reply {
			struct _sysmsg_hdr_reply hdr;
			struct _sysmsg_version_reply version;
			} ver_reply;
		} *m;

	m = buf;

	switch(m->ver.hdr.subtype) {

	case _SYSMSG_SUBTYPE_VERSION:
		memset(&m->ver_reply, 0, sizeof(m->ver_reply));
		m->ver_reply.hdr.status = 0;
		strcpy(&m->ver_reply.version.name[0], "Dev.x25");
		strncpy(&m->ver_reply.version.date[0], __DATE__, 11);
		m->ver_reply.version.version = 424;
		m->ver_reply.version.letter = 'A';
		return(sizeof(m->ver_reply));

	case _SYSMSG_SUBTYPE_DEATH:
	case _SYSMSG_SUBTYPE_TRACE:
	default:
		m->status = EINVAL;
		return(sizeof(m->status));
		}
	}

/*
 * Process all input data in all devices
 * Only return when all input is exhausted.
 * Rely on a subsequent Interrupt (which triggers 'in_proxy') to wake us up.
 */

int input_ready() {
	pid_t proxy;
	int i, n;
	char buf[128];

	proxy = 0;

	x25receive();

	/*
	 * Scan all devices looking for available data
	 */
	for( i = 0; i < num_devices; ++i ) {
		while( (n = get_data(i, &buf[0], sizeof(buf))) > 0 ) {
			if(drv_ctrl.ttim_offset != 0) {
				/*
				 * Use multiple character input routine if supported by Dev
				 */
				pid_t (far *ttimcall) (int tty, int len, char far *data, unsigned ds);
				ttimcall = MK_FP( FP_SEG(drv_ctrl.tti), drv_ctrl.ttim_offset );
				proxy |= (*ttimcall) (ctrl[i].tty, n, &buf[0], drv_ctrl.tti_ds);
				}
			else {
				/*
				 * Must inject character at a time for older Dev's
				 */
				int j;
				for(j = 0; j < n; ++j)
					proxy |= (*drv_ctrl.tti) (buf[j], ctrl[i].tty, drv_ctrl.tti_ds);
				}
			}
		}

	/*
	 * "wake up" Dev if ANYTHING of interest happened
	 */
	if(proxy)
		Trigger(proxy);

	return(0);
	}

void urgentwork( void )
{
	int u;

	if ( urgent==-1 ) return;

	do {
	urgent = -1;

	for( u=0; u<MAX_DEVICES; u++ )
		{
		if ( pkt_urgent[u] & URGENT_LIB )
			{
			int cv = unit2cv( u );
			if ( debug )
				fprintf( stderr, "urgentwork() LIB unit=%d cv=%d\n", u, cv );
			libcv( cv, 0, 0 );
			pkt_urgent[u] -= URGENT_LIB;
			}
		if ( pkt_urgent[u] & URGENT_ECHO_ON )
			{
			int cv = unit2cv( u );
			if ( debug )
				fprintf( stderr, "urgentwork() ECHO_ON unit=%d cv=%d\n", u, cv );
			echocv( cv, 1 );
			pkt_urgent[u] -= URGENT_ECHO_ON;
			}
		if ( pkt_urgent[u] & URGENT_ECHO_OFF )
			{
			int cv = unit2cv( u );
			if ( debug )
				fprintf( stderr, "urgentwork() ECHO_OFF unit=%d cv=%d\n", u, cv );
			echocv( cv, 0 );
			pkt_urgent[u] -= URGENT_ECHO_OFF;
			}
		}

	/* urgent may be updated by Dev32 under IT */
	} while( urgent != -1 );
}

/*
 * This function will print forever, only returning
 * when all output buffers are drained
 * Rely on a subsequent Kick() call to wake us up again
 *
 * NOTE: that the memory semaphores 'idle' and 'more' are updated
 *       by the far-call stubs (ie: Kick) to make this work efficiently
 */

void flush_output( void ) {
	struct output_buffer far *obuf;
	char far *tail;
	char far *end;
	char far *begin;
	int i, n, size;
	int nothing, unit;
	int expand_newlines;

	urgentwork();

	for(;;) {

		/* Scan every output queue, looking for something to do */
		nothing = 0;
		more = 0;
		for(unit = 0; unit < num_devices; ++unit) {

			x25receive();

			if ( pkt_indisponible == 1 )
				{
				++nothing;
				continue;
				}

			if ( pkt_cv_saturer[unit] == 1 )
				{
				++nothing;
				continue;
				}
	
			obuf = ctrl[unit].obuf;

			/*
			 * Process flow control output functions first
			 */

			if(obuf->mode & OUTPUT_HFLOW_OFF) {
				obuf->mode &= ~OUTPUT_HFLOW_OFF;
				drop_hw(unit);
				}
	
			if(obuf->mode & OUTPUT_HFLOW_ON) {
				obuf->mode &= ~OUTPUT_HFLOW_ON;
				raise_hw(unit);
				}

			if(obuf->mode & OUTPUT_XOFF) {
				obuf->mode &= ~OUTPUT_XOFF;
				/*
				 * Disable input by sending XOFF
		   	 	 */
				put_data(unit, (char far *) "\023", 1, 0);
				}
	
			if(obuf->mode & OUTPUT_XON) {
				obuf->mode &= ~OUTPUT_XON;
				/*
				 * Re-enable input by sending XON
			 	 */
				put_data(unit, (char far *) "\021", 1, 0);
				}
	
			/* Skip if no data characters available */

			_disable();
			if(obuf->size == 0  ||
					(obuf->mode & (OUTPUT_STOPPED|OUTPUT_STOPPED_HW)) != 0) {
				_enable();
				++nothing;
				continue;
				}
			_enable();

			if(obuf->discard) {
				/*
				 * Discard some (or all) data instead of displaying it
				 */
				n = obuf->discard;
				_disable();
				obuf->discard -= n;
				obuf->size -= n;
				_enable();
				i = &obuf->buffer[obuf->length] - obuf->tail;
				if(i > n)
					obuf->tail += n;
				else
					obuf->tail = &obuf->buffer[n-i];
				}

			/* Extract the data from the buffer */

			tail = obuf->tail;
			end = &obuf->buffer[obuf->length];
			begin = &obuf->buffer[0];
			size = obuf->size;

			expand_newlines = obuf->mode & OUTPUT_XNL;

			/* Calculate number of chars before wrap-around point */
			i = end - tail;
			if(i >= size) { /* No wrap around */
				n = put_data(unit, tail, size, expand_newlines);
				}
			else {
				n = put_data(unit, tail, i, expand_newlines);

				// ATTENTION commentaire obligatoire  cf: pkt_cv_saturer[]
				// n += put_data(unit, begin, size - i, expand_newlines);
				}

			/*
			 * Update the Queue pointers, now that the
			 * 'n' bytes of data has been extracted
			 */

			_disable();

			if(i > n) /* No wrap around */
				obuf->tail = tail + n;
			else
				obuf->tail = begin + (n-i);
			obuf->size -= n;
			/*
			 * Compensate for any flushed data which
			 * was in the process of being transmitted
			 */
			if(obuf->discard > n)
				obuf->discard -= n;
			else
				obuf->discard = 0;

			_enable();
	
			/*
			 * If size of buffer drops below notification threshold,
			 * inform DEV of the new information
			 */
			_disable();
			if(obuf->size < obuf->bcount) {
				obuf->bcount = 0;
				obuf->action |= ACT_OUTPUT_READY;
				_enable();
				Trigger(drv_ctrl.hw_pid);
				}
			else
				_enable();
			}

		/*
		 */
		_disable();
		if(nothing >= num_devices  &&  more == 0) {
			idle = 1;
			_enable();
			return;
			}
		else
			_enable();
		}
	}

int init_driver()
	{
	if ( (fdasmt = open( devasmt, O_RDWR )) < 0 )
		perror( "init_driver() open" );

	fcntl( fdasmt, F_SETFD, FD_CLOEXEC);

	old_mode = dev_mode( fdasmt, 0, _DEV_MODES);
	dev_arm( fdasmt, in_proxy, _DEV_EVENT_RXRDY );

	return(0);
	}

int exit_driver()
	{
	dev_mode( fdasmt, old_mode, _DEV_MODES);
	return(0);
	}

get_data(int unit, char *buf, int nbytes)
	{
	/*
	 * Get (at most) 'nbytes' of data from 'unit' (into 'buf').
	 * Return #bytes read.
	 */

	int l = strlen( xqueue[unit] );

	if ( l == 0 )
		return( 0 );

	if ( l < nbytes )
		{
		memcpy( buf, xqueue[unit], l );
		memset( xqueue[unit], 0, XQSIZE );
		return( l );
		}
	else
		{
		memcpy( buf, xqueue[unit], nbytes );
		memcpy( xqueue[unit], xqueue[unit]+nbytes, l-nbytes );
		memset( xqueue[unit]+l-nbytes, 0, XQSIZE-(l-nbytes) );
		return( nbytes );
		}
	}

put_data(int unit, char far *buf, int nbytes, int expand_newlines)
	{
	/*
	 * Output 'nbytes' of data (in 'buf') to 'unit'.
	 * Return #bytes written (before expansion).
	 */

	/* Your stuff here */
	int i, j, r, n = nbytes>127 ? 127:nbytes;
	char lbuf[ 256 ];

	if ( debug )
		fprintf( stderr, "Dev.x25: put_data( unit=%d, buf, nbytes=%d, exp=%d ) n=%d unit2cv(unit %d)=%d\n",
			unit, nbytes, expand_newlines, n, unit, unit2cv(unit) );

	/* solve, cv_saturer & indisponible */
	x25receive();

	if ( pkt_cv_utiliser[ unit ]==0 )
		{
		if ( debug )
			fprintf( stderr, "Dev.x25: WARNING put_data() nbytes=%d pkt_cv_utiliser[ unit=%d ]==0 FALSE\n",
				nbytes, unit );
		return( nbytes );
		}
	if ( pkt_cv_saturer[ unit ]==1 )
		{
		if ( debug )
			fprintf( stderr, "Dev.x25: WARNING put_data() nbytes=%d pkt_cv_saturer[ unit=%d ]==1 TRUE\n",
				nbytes, unit );
		return( 0 );
		}
	if ( pkt_indisponible==1 )
		{
		if ( debug )
			fprintf( stderr, "Dev.x25: WARNING put_data() nbytes=%d pkt_indisponible==1 TRUE\n",
				nbytes );
		return( 0 );
		}

	memset( lbuf, 0, sizeof(lbuf) );
	for( i=j=0; i<n && j<127; i++,j++ )
		if ( buf[i]=='\n' && expand_newlines )
			{
			lbuf[j++] = '\r';
			lbuf[j] = '\l';
			}
		else
			lbuf[j] = buf[i];
	n = j;
	r = i;

	xbuf[0] = STX;
	xbuf[1] = 0x30;
	xbuf[2] = unit2cv(unit);
	xbuf[3] = 0;
	xbuf[4] = n;
	for( i=0; i<n; i++ )
		xbuf[ 5+i ] = lbuf[i];
	xbuf[ n + 5 ] = ETX;
	write( fdasmt, xbuf, n+6 );

	if ( debug>1 )
		dumper( xbuf, n+5, stderr );

	return( r );
}

void drop_hw(int unit) {
	/*
	 * Disable input by dropping Hardware handshaking
 	 */

	if ( debug )
		fprintf( stderr, "drop_hw(unit=%d)\n", unit );
	}

void raise_hw(int unit) {
	/*
	 * Re-enable input by raising Hardware handshaking
 	 */

	if ( debug )
		fprintf( stderr, "raise_hw(unit=%d)\n", unit );
	}


void x25receive()
{
	int n, type, cv, lng, u, l, i;

	urgentwork();

	//
	// Le format d'une trame est,  STX TM NT CL1 CL0 .... ETX
	//                             1er 2e 3e  4e  5e 

re:
	n = dev_ischars( fdasmt );
	if ( n < 1 )
		{
		dev_arm( fdasmt, in_proxy, _DEV_EVENT_RXRDY );
		return;
		}

	memset( xbuf, 0, sizeof(xbuf) );

	dev_read( fdasmt, xbuf, 1, 1, 0, 0, 0, 0);
	if ( xbuf[0] != STX )
		{
		fprintf( stderr, "Dev.x25: err STX, rcv 0x%02x", xbuf[0] );
		dev_arm( fdasmt, in_proxy, _DEV_EVENT_RXRDY );
		return;
		}

	dev_read( fdasmt, xbuf, 4, 4, 0, 0, 0, 0);
	type = xbuf[0];
	cv = xbuf[1];
	lng = (int) (xbuf[2] & 0x7f) * 128 + (int) (xbuf[3] & 0x7f);

	if ( debug )
	fprintf( stderr, "Dev.x25:  type %02x  cv %02x  lng %3d (%02x %02x)\n",
		type, cv, lng, xbuf[2], xbuf[3] );

	if ( lng )
		{
		memset( xbuf, 0, sizeof(xbuf) );
		dev_read( fdasmt, xbuf, lng, lng, 0, 0, 0, 0);
		if ( debug )
			dumper( xbuf, lng, stderr );
		}

	switch( type )
		{

		/* Trame d'information */
		case 0x30:
					if ( debug )
						fprintf( stderr, "Trame INFO\n" );

					u = cv2unit( cv );
					/* this doesn't allow 0x00 in incoming queue */
					for( l=strlen(xqueue[u]),i=0; i<lng && (i+l)<XQSIZE; i++ )
						xqueue[u][l+i] = xbuf[i];
					break;

		/* Trame d'appel */
		case 0x31:
					if ( debug )
						fprintf( stderr, "Trame APPEL\n" );

					acceptation( cv, xbuf, lng );

					u = cv2unit( cv );
					pkt_cv_utiliser[ u ] = 1;
					pkt_cv_saturer [ u ] = 0;
					pkt_cv_echo [ u ] = 1;

					memset( xqueue[u], 0, XQSIZE );
	
					if ( debug )
						fprintf( stderr, "RCV_CARRIER unit=%d\n", u );

					i = (*drv_ctrl.tti) ( RCV_CARRIER, ctrl[u].tty, drv_ctrl.tti_ds);
					if ( i )
						Trigger( i );

					break;

		/* Trame de liberation */
		case 0x32:
					if ( debug )
						fprintf( stderr, "Trame LIBE\n" );

					u = cv2unit( cv );
					pkt_cv_utiliser[ u ] = 0;
					if ( pkt_cv_saturer[u]==1 )
						Trigger( out_proxy );
					pkt_cv_saturer [ u ] = 0;
					pkt_cv_echo [ u ] = 1;

					memset( xqueue[u], 0, XQSIZE );
	
					if ( debug )
						fprintf( stderr, "RCV_HNGUP unit=%d\n", u );

					i = (*drv_ctrl.tti) ( RCV_HNGUP, ctrl[u].tty, drv_ctrl.tti_ds);
					if ( i )
						Trigger( i );

					break;

		/* Trame incoherence dans donnees */
		case 0x3f:
					if ( debug )
						fprintf( stderr, "Recu Trame INCO\n");
					break;

		/* Trame d'indication d'erreur de procedure de multiplexage */
		case 0x34:
					if ( debug )
						fprintf( stderr, "Recu Trame ERR Multiplexage\n");
					break;

		/* Trame d'indisponibilite / Saturation d'un CV */
		case 0x35:
					if ( debug )
						fprintf( stderr, "Recu Trame INDISPONIBLE (saturation) cv=%d unit=%d\n",
							cv, cv2unit(cv) );
					
					pkt_cv_saturer [ cv2unit(cv) ] = 1;
					break;

		/* Trame de disponibilite / Fin de saturation */
		case 0x36:
					if ( debug )
						fprintf( stderr, "Recu Trame DISPO\n");

					pkt_cv_saturer[ cv2unit(cv) ] = 0;
					Trigger(out_proxy);
					break;

		/* Trame Concentrateur Indisponible */
		case 0x39:
					if ( debug )
						fprintf( stderr, "Trame Concentrateur KPUT (Attendre)\n");

					pkt_indisponible = 1;
					break;

		/* Trame Concentrateur Disponible */
		case 0x40:
					if ( debug )
						fprintf( stderr, "Trame Concentrateur OK (Va-Bene)\n");

					pkt_indisponible = 0;
					Trigger(out_proxy);
					break;

		/* Trame de donnees avec le bit Q a 1 */
		case 0x3a:
					if ( debug )
						fprintf( stderr, "Trame DATA BIT Q\n");
					break;

		/* Trame de suppression d'echo */
		case 0x37:
					if ( debug )
						fprintf( stderr, "ATTENTION: Recu Trame (ECHO_OFF) ???\n");
					break;

		/* Trame de mise d'echo */
		case 0x38:
					if ( debug )
						fprintf( stderr, "ATTENTION: Recu Trame (ECHO_ON) ???\n");
					break;

		/* Trame de Reset de du concentrateur */
		case 0x3e:
					if ( debug )
						fprintf( stderr, "ATTENTION: Recu Trame 0x3e (RESET) ???\n");
					break;

		/* Trame Communication etablie */
		case 0x33:
					if ( debug )
						fprintf( stderr, "ATTENTION: Recu Trame 0x33 (OKCOM) ???\n");
					break;
		default:
					if ( debug )
						fprintf( stderr, "Xadm: x25reveive() WARNING !!!! unknown message type=0x%02x\n", type );
					break;
		}

	dev_read( fdasmt, xbuf, 1, 1, 0, 0, 0, 0);
	if ( xbuf[0] != ETX )
		fprintf( stderr, "Dev.x25: err ETX, rcv 0x%02x\n", xbuf[0] );

	// parse all input in one shot
	goto re;
}

void acceptation( unsigned char vc, char *data, int lng )
{
	char xwbuf[16];
	int cpid, fd, i;
	static char iov[10];
	static char *larg[10];
	static char bdat[256];

	if ( debug )
		fprintf( stderr, "acceptation: cv(%02d) lng(%d) data(%s) tty(%d)\n",
			vc, lng, data, cv2tty(vc) );

	/* Trame d'acception d'un CV */
	xwbuf[0] = STX;
	xwbuf[1] = 0x33;
	xwbuf[2] = vc;
	xwbuf[3] = 0x00;
	xwbuf[4] = 0x00;
	xwbuf[5] = ETX;

	write( fdasmt, xwbuf, 6 );

	/* start login with stdin,stdout,stderr redirected on x25 tty */

	sprintf( xwbuf, "/dev/ttx%d", cv2tty(vc) );
	larg[0] = "login";
	larg[1] = NULL;
	memset( iov, -1, sizeof(iov) );
	fd = open( xwbuf, O_RDWR );
	if ( fd < 0 ) return;

	fcntl( fd, F_SETFD, FD_CLOEXEC);

	memset( bdat, 0, sizeof(bdat) );
	for( i=0; i<lng && i<sizeof(bdat); i++ )
		if ( data[i] != 0 )
			bdat[i] = data[i];
		else
			bdat[i] = ' ';

	larg[0] = "login";
	larg[1] = &bdat[0];
	larg[2] = NULL;
	memset( iov, -1, sizeof(iov) );

	fcntl( fd, F_SETFD, FD_CLOEXEC);
	iov[0] = iov[1] = iov[2] = fd;

	errno = 0;
	cpid = qnx_spawn(	0,
				NULL,
				getnid(),
				10, 2,
					_SPAWN_SIGCLR | 
					_SPAWN_SETSID | 
					_SPAWN_TCSETPGRP | 
					_SPAWN_NOZOMBIE,
				startprog,
				larg,
				environ,
				iov,
				0	);

	close( fd );

	if ( debug )
		fprintf( stderr, "qnx_spawn() cpid = %d, errno = %d\n", 
			cpid, errno );
}

void libcv( int vc, int cause, int diag )
{
	char xwbuf[ 16 ];
	int u, i;

	xwbuf[0] = STX;
	xwbuf[1] = 0x32;
	xwbuf[2] = vc;
	xwbuf[3] = 0;
	xwbuf[4] = 3;
	xwbuf[5] = cause;
	xwbuf[6] = 0x2f;
	xwbuf[7] = diag;
	xwbuf[8] = ETX;

	write( fdasmt, xwbuf, 9 );

	u = cv2unit(vc);

	if ( debug )
		fprintf( stderr, "libcv( cv=%02x, cause=%d, diag=%d ) unit=%d\n",
			vc, cause, diag, u );

	pkt_cv_utiliser[ u ] = 0;
	pkt_cv_echo[ u ] = 1;

	/* Trash, output data */
	if ( pkt_cv_saturer[u] == 1 )
		Trigger( out_proxy );

	pkt_cv_saturer[ u ] = 0;

	i = (*drv_ctrl.tti) ( RCV_HNGUP, ctrl[u].tty, drv_ctrl.tti_ds);
	if ( i )
		Trigger( i );
}

void echocv( int vc, int on )
{
	char xwbuf[ 16 ];
	int u = cv2unit(vc);

	if ( pkt_cv_utiliser[u] == 0 ) return;
	if ( pkt_cv_echo[u]==1 && on ) return;
	if ( pkt_cv_echo[u]==0 && !on ) return;

	if( debug )
		fprintf( stderr, "echo_cv( cv=%d, on=%d ) unit=%d\n", vc, on, u );

	xwbuf[0] = STX;
	xwbuf[1] = on? 0x38:0x37;
	xwbuf[2] = vc;
	xwbuf[3] = 0;
	xwbuf[4] = 0;
	xwbuf[5] = ETX;

	write( fdasmt, xwbuf, 6 );

	pkt_cv_echo[u] = on;
}

void dumper( buffer, lng, fd )
char *buffer;
int lng;
FILE *fd;
{
	register int	i;
	int				p;

	if ( debug < 2 ) return;

	for ( p = 0; p < lng; p += 16 )
	{
	for ( i = 0; i < 16; i++ )
		fprintf( fd, "%02x ", (unsigned)(char)buffer[ i + p ] );

	fprintf ( fd, "| " );

	for ( i = 0; i < 16; i++ )
		fprintf( fd, "%c",
			( (unsigned)(char)buffer[ i + p ] > 31 ) ?
			( (unsigned)(char)buffer[ i + p ] < 0x80 ) ?
				buffer[ i + p ] : '.' : '.' );

	fprintf ( fd, "\n" );
	}
}

/* 
 * Far call interface routines.
 *
 * These routines are far called by Dev which requires
 * them to be compiled specially. In particular, it is very important
 * that they be compiled with -zu (SS != DS) since this will be
 * the case. Also stack checking should be disabled
 *
 * Kick() is called by Dev whenever "something" has changed in the
 *        output buffer (new data, change in flow control, etc.)
 *
 * Stty() is called by Dev when something "significant" has changed
 *        in the termios control structure which requires hardware
 *        attention (eg: baud rate, parity).
 *
 * Ctrl() is called to perform various "houskeeping" functions
 *        (like dropline by rasining/lowering DTR, sendbreak, etc.)
 * 
 * Inside these routines, teh following limitations apply:
 *
 * 1) No system calls other than Trigger() can be used. Since most
 *    C library routines use system calls internally (eg: read
 *    will use Sendfd), the "safe" rule is to avoid ALL library
 *    routines completetly. The best policy is to only access/change
 *    global memory variables and Trigger() a proxy to wake up your
 *    process if more work needs to be done.
 *
 * 2) No mapped or segmented memory can be accesses (ie: no far pointers)
 *    since Dev's process context will not have access to them. Only
 *    variables in DS can be accessed.
 *
 * 3) Try to avoid much processing since Dev (at high priority) will be
 *    effectively blocked until your routine finishes.
 *
 * 4) Don't use much stack (keep it under a few hundred bytes).
 */

/*
 *	declare the calls 'dev' will make into driver.
 */
int far	Kick  (struct output_buffer far *obuf);
int far	Stty  (struct termios far *tios); 
int far	Ctrl  (union ioctl_entry far *ioc); 

#pragma aux (_dev_call) Kick;
#pragma aux (_dev_call) Stty;
#pragma aux (_dev_call) Ctrl;

#pragma off (check_stack);

/*
 * Kick is called by DEV everytime something "new" happens
 * to the output buffer
 *
 * It will set 'obuf' to be the output buffer
 * It will also set ds to be the data segment of the driver being called
 *
 * Since this routine is "called" by DEV, it will ALWAYS run at higher
 * priority than the driver task, so cannot be interrupted.
 */

int far Kick (struct output_buffer far *obuf)	{
	extern int idle;
	extern int more;
	extern pid_t out_proxy;

	if((obuf->size > 0
			&&  (obuf->mode & (OUTPUT_STOPPED|OUTPUT_STOPPED_HW)) == 0)
			||  (obuf->mode & (OUTPUT_HFLOW_OFF|OUTPUT_HFLOW_ON|OUTPUT_XON|OUTPUT_XOFF)) ) {
		if(idle) {
			idle = 0;
			Trigger(out_proxy);
			}
		else
			more = 1;
		return(1);
		}
	return(0);
	}


/*
 * Stty is called by DEV to indicate a significant change in the
 * termios control structure (baud rate, modem control)
 * It can optionally be called for EVERY call to tcsetattr() if the
 * DRVR_STTY option os passed to dev_drvr_register().
 *
 * It will set 'tios' to be a far pointer to the termios entry
 * It will also set ds to be the data segment of the driver being called
 *
 * Since this routine is "called" by DEV, it will ALWAYS run at higher
 * priority than the driver task, so cannot be interrupted.
 */

int far Stty (struct termios far *tios)
	{

	/* En X25, echo du PAVI (minitel) */

	if ( tios->c_lflag & ECHO )
		tios->c_lflag -= ECHO;

	return(0);
	}


/*
 * Ctrl is called by DEV to indicate a high level
 * control function
 *
 * It will set 'ioc' to be an ioctl control block
 * It will also set ds to be the data segment of the driver being called
 *
 * Since this routine is "called" by DEV, it will ALWAYS run at higher
 * priority than the driver task, so cannot be interrupted.
 */

int far Ctrl (union ioctl_entry far *ioc)
	{
	int type, u;

	type = ioc->type;
	ioc->status = IOCTL_ERROR;

	switch (type) {

	case IOCTL_OPEN:
		/*
		 * First open on device
		 */
		ioc->status = IOCTL_OK;
		return(0);

	case IOCTL_CLOSE:
		/*
		 * Last close on device
		 */
		ioc->status = IOCTL_OK;

		/* If there's no process on the CV, and CV is marked used,
           free this CV as soon as possible */

		u = ioc->close.unit-1;
		if ( pkt_cv_utiliser[ u ]==1 )
			{
			urgent = 1;
			pkt_urgent[ u ] |= URGENT_LIB;
			Trigger( in_proxy );
			}

		return(0);

	case IOCTL_CTL:	
		ioc->user_return.status = IOCTL_OK;

		u = ioc->user.unit-1;
		if ( pkt_cv_echo[ u ]==0 && ioc->user.data[0]==1 )
			{
			urgent = 1;
			pkt_urgent[ u ] |= URGENT_ECHO_ON;
			Trigger( in_proxy );
			}
		if ( pkt_cv_echo[ u ]==1 && ioc->user.data[0]==2 )
			{
			urgent = 1;
			pkt_urgent[ u ] |= URGENT_ECHO_OFF;
			Trigger( in_proxy );
			}
		return(0);

	case IOCTL_STARTDROP:
		ioc->status = IOCTL_OK;

		/* If we get drop line, tcdropline() from used CV,
           free this CV, and SIGHUP process */

		u = ioc->startdrop.unit-1;
		if ( pkt_cv_utiliser[ u ]==1 )
			{
			urgent = 1;
			pkt_urgent[ u ] |= URGENT_LIB;
			Trigger( in_proxy );
			}

		return(0);

	case IOCTL_STARTBREAK:
	case IOCTL_STOPBREAK:
	case IOCTL_STOPDROP:
		ioc->status = IOCTL_OK;
		return(0);

		}

	return(-1);
	}

#pragma on (check_stack);
